#include <unistd.h>
#include "fm_keyboard.h"
#include "fm_conf.h"
#include "sc_action.h"
#include "fm_tree.h"
#include "sc_file.h"
#include "fm_vfs.h"
#include "fm_menu.h"
#include "fm_defs.h"
#include "sc_archive.h"
#include "stopwatch.h"
#include "fm_helpscreens.h"
#include "sc_shell.h"

#ifndef _TARGET_MINGW_
  #include <sys/vfs.h>
#endif
string lastdestination="";

bool _progress_by_count;

cWindow* win;
long totalfilesize,partfilesize;
long totalfilecount,partfilecount;
cStopwatch timer;

//Prototype
void ShowProgress(long int thispart, long int thistotal, int justdone);

/*************************************************************************/
//ACTIONS
/*************************************************************************/

class cCopyAction: virtual public cBaseAction {
  public:
    cCopyAction(const cDirEntry& de, const string& sname,const string& dname,tCheckAction ovac, tProgFunc progress)
      : cBaseAction(de,sname,progress) {
      destname=dname;
      overwriteaction=ovac;
    }
    bool Action() {return InternalCopyAction();};
  protected:
    string destname;
    tCheckAction overwriteaction;
    bool InternalCopyAction() {
      #ifdef _SYMLINKS_EXIST_
        if (entry.islink) {
          if (destname!="") {
            return (symlink(entry.linkto.c_str(),destname.c_str())!=-1);
          }
          return false; //later, later
        }
      #endif
      if (MatchStart(_PRINTERTOKEN,destname)) {
        return PrintFile(entry.fullname);
      }

      if (entry.isdir) {
        return MkDir(destname);
      }
      cFile* f1=GetFilePointer(entry.fullname);
      if (!f1->Open(omRead)) {delete f1; return false;};

      cFile* f2=GetFilePointer(destname);
      if (!f2->Open(omWrite)) {f1->Close(); delete f1; delete f2; return false;};

      char buf[65536];
      int i=f1->Read(&buf,sizeof(buf));
      while (i>0) {
        f2->Write(&buf,i);
        progfunc(f1->Pos(),f1->Size(),i);
        i=f1->Read(&buf,sizeof(buf));
      }
      f2->SetMode(f1->GetMode());
      f1->Close(); f2->Close();
      delete f1; delete f2;
      return true;
    }
};

class cEraseAction : virtual public cBaseAction {
  public:
    cEraseAction(const cDirEntry& de, const string& sname, tProgFunc progress)
      : cBaseAction(de,sname,progress) {
    }
    bool Action() {return InternalEraseAction();};
  protected:
    bool InternalEraseAction(bool doprogress=true) {
      if (doprogress) progfunc(2,2,entry.size);
      if (entry.isdir) {
        return RmDir(entry.fullname);
      } else {
        return DeleteFile(entry.fullname);
      }

    }
};

class cMoveAction : public cCopyAction, public cEraseAction {
  public:
    cMoveAction(const cDirEntry& de, const string& sname,const string& dname,tCheckAction ovac, tProgFunc progress)
      : cCopyAction(de,sname,dname,ovac,progress),
        cEraseAction(de,sname,progress),
        cBaseAction(de,sname,progress) {
    }
    bool Action() {return InternalMoveAction();};
  protected:
    bool InternalMoveAction() {
      //Try the fast way
      if (Rename(entry.fullname,destname)) {
        progfunc(2,2,entry.size);
        return true;
      }
      //Well, we have to actually move data
      if (InternalCopyAction()) {
        return InternalEraseAction(false);
      } else return false;
    }
};

class cUnpackAction : public cBaseAction {
  public:
    cUnpackAction(const cDirEntry& de, const string& sname,const string& dname, tProgFunc progress)
      : cBaseAction(de,sname,progress) {
      destdir=dname;
    }
    bool Action() {return InternalUnpackAction();};
  protected:
    string destdir;
    bool InternalUnpackAction() {
      return (UnpackWholeFile(entry.fullname, destdir, progfunc));
    }
};


class cChmodeAction : public cBaseAction {
  public:
    cChmodeAction(const cDirEntry& de, const string& sname, sModeChanges modechange, tProgFunc progress)
      : cBaseAction(de,sname,progress) {
      modch=modechange;
      usemode=true;
    }
    cChmodeAction(const cDirEntry& de, const string& sname, sAttrChanges attrchange, tProgFunc progress)
      : cBaseAction(de,sname,progress) {
      attrch=attrchange;
      usemode=false;
    }
    bool Action() {return InternalChmodeAction();};
  protected:
    union {
      sModeChanges modch;
      sAttrChanges attrch;
    };
    bool usemode;
    bool InternalChmodeAction() {
      if (usemode) {
        sModeChanges md=TranslateMode(entry.modeattr);
        if (modch.row!=_nochange) md.row=modch.row;
        if (modch.wow!=_nochange) md.wow=modch.wow;
        if (modch.xow!=_nochange) md.xow=modch.xow;

        if (modch.rgr!=_nochange) md.rgr=modch.rgr;
        if (modch.wgr!=_nochange) md.wgr=modch.wgr;
        if (modch.xgr!=_nochange) md.xgr=modch.xgr;

        if (modch.rot!=_nochange) md.rot=modch.rot;
        if (modch.wot!=_nochange) md.wot=modch.wot;
        if (modch.xot!=_nochange) md.xot=modch.xot;

        if (modch.suid!=_nochange) md.suid=modch.suid;
        if (modch.sgid!=_nochange) md.sgid=modch.sgid;
        if (modch.sticky!=_nochange) md.sticky=modch.sticky;
        progfunc(2,2,entry.size);  //updateprogress
        return SetFileMode(entry.fullname,md);
      } else {
      #ifdef _TARGETFLAVOR_DOS_
        sAttrChanges at=TranslateAttr(entry.modeattr);
        if (attrch.hidden!=_nochange) at.hidden=attrch.hidden;
        if (attrch.readonly!=_nochange) at.readonly=attrch.readonly;
        if (attrch.system!=_nochange) at.system=attrch.system;
        if (attrch.archive!=_nochange) at.archive=attrch.archive;
        progfunc(2,2,entry.size); //Update progress
        return SetFileAttr(entry.fullname,at);
      #else
        return false;
      #endif
      }
    }
};


/*************************************************************************/
//ACTION PROJECT
/*************************************************************************/


//General initialisation for both single and multible files
void cActionProject::Initialize() {
  breaked=allpressed=error=false;
  filecount=filemeasure=totalmeasure=failures=0;
  howto_subdirs=howto_overwrite=howto_failure=oaUndetermined;
  #ifdef _SYMLINKS_EXIST_
    howto_symlinks=oaUndetermined;
  #endif
  destdir="";
  error=!SetupAction();
  string s;
  switch (atype) {
    case acCopy :
    case acMove : _progress_by_count=false; break;
    default : _progress_by_count=true;
  }
  w=NULL;
  if (!error) {
    w=new cWindow(ScreenW()-10,ScreenH()-5,"Scanning..."); w->ChUpright("~Yes/No/All/Quit");
    win=w; //Ush, bad globals, but it works... For communication with Progress function
  }
}

/*
//Do action on a single file
cActionProject::cActionProject(const cDirEntry& singlefile, const string& dir, tActionType atyp) {
  atype=atyp; fromdir=AddSlash(dir);
  Initialize();
  if (error) return;
  #ifdef _SYMLINKS_EXIST_
    if (singlefile.islink) {
      AskSymlinkAction();
      if (howto_symlinks==oaIgnore) return;
      if (howto_symlinks==oaCopy) {
        RegisterFile(singlefile);
        return;
      }
    }
  #endif
  howto_subdirs=oaAlways; //if you choose a single dir, the you probably want to
                          //recurse, right?
  if (singlefile.isdir) {
    if (atype==acAttr) {
      if (howto_subdirs==oaUndetermined) AskDirAction();
    }
    switch (atype) {
      case acErase : break;
      case acMove : RegisterFile(singlefile,acCopy); break;
      default : RegisterFile(singlefile);
    }
    ScanDir(singlefile.fullname);
    switch (atype) {
      case acErase :
      case acMove : RegisterFile(singlefile,acErase); break;
      default : {;};
    }
  } else {
    RegisterFile(singlefile);
  }
}*/

cActionProject::cActionProject(const tDirContainer& filelist, const string& dir, tActionType atyp) {
  atype=atyp; fromdir=AddSlash(dir);
  Initialize();
  if (error) return;
  RegisterFilelist(filelist);
}

cActionProject::~cActionProject() {
  scDELETE(w);
  if (failures>0) {
    ErrorBox(MkStr(failures)+" failures");
  }
}



/***************************************************************************/

//Just a small helperfunction
string as2str(int as) {
  switch (as) {
    case _set : return "Set";
    case _clear : return "Clear";
    default : return "NoChange";
  }
}

//Dialog to allow user to change filemode on unixlike filesystems
bool cActionProject::SetModeChanges() {
  modech.row=modech.wow=modech.xow=_nochange;
  modech.rgr=modech.wgr=modech.xgr=_nochange;
  modech.rot=modech.wot=modech.xot=_nochange;
  modech.suid=modech.sgid=modech.sticky=_nochange;
  tKey k;
  bool changed=false;
  do {
    //short int user=0, group=0, other=0, special=0;

    cWindow w(60,15,"Mode Changes");
    w.Write(1,1,"Permissions:",chRED);

    w.Write(1,2,"Owner              Group               Others",chYELLOW);
    w.Write(1,3,"1 Read : "+as2str(modech.row)); //if (modech.row==_set) user+=4;
    w.Write(1,4,"2 Write: "+as2str(modech.wow)); //if (modech.wow==_set) user+=2;
    w.Write(1,5,"3 Exec : "+as2str(modech.xow)); //if (modech.xow==_set) user+=1;

    w.Write(20,3,"4 Read : "+as2str(modech.rgr)); //if (modech.rgr==_set) group+=4;
    w.Write(20,4,"5 Write: "+as2str(modech.wgr)); //if (modech.wgr==_set) group+=2;
    w.Write(20,5,"6 Exec : "+as2str(modech.xgr)); //if (modech.xgr==_set) group+=1;

    w.Write(40,3,"7 Read : "+as2str(modech.rot)); //if (modech.rot==_set) other+=4;
    w.Write(40,4,"8 Write: "+as2str(modech.wot)); //if (modech.wot==_set) other+=2;
    w.Write(40,5,"9 Exec : "+as2str(modech.xot)); //if (modech.xot==_set) other+=1;

    w.Write(1,7,"Specials:",chRED);
    w.Write(1,8, "A Set ~UserID     : "+as2str(modech.suid));    //if (modech.suid==_set) special+=4;
    w.Write(1,9, "B Set ~GroupID    : "+as2str(modech.sgid));    //if (modech.sgid==_set) special+=2;
    w.Write(1,10,"C Set ~Sticky bit : "+as2str(modech.sticky));  //if (modech.sticky==_set) special+=1;

    w.Write(30,7,"Shortcuts:",chRED);
    w.Write(30,8, "D Set Exec-bit for all");
    w.Write(30,9, "E Set Write-bit for all");
    w.Write(30,10,"F Clear Write-but for all but");
    w.Write(30,11,"  owner");

    /*string s;
    s+=(MkStr(special)+MkStr(user)+MkStr(group)+MkStr(other));
    w.Write(1,12,"Similar commandline: chmod "+s+" [name]",chYELLOW);*/

    w.Write(1,14,"Enter to activate  ESC/F10 to abort",chYELLOW);

    k=GetSomeKey();
    switch (k) {
      case k1 : (modech.row==2 ? modech.row=0 : modech.row++); changed=true; break;
      case k2 : (modech.wow==2 ? modech.wow=0 : modech.wow++); changed=true; break;
      case k3 : (modech.xow==2 ? modech.xow=0 : modech.xow++); changed=true; break;

      case k4 : (modech.rgr==2 ? modech.rgr=0 : modech.rgr++); changed=true; break;
      case k5 : (modech.wgr==2 ? modech.wgr=0 : modech.wgr++); changed=true; break;
      case k6 : (modech.xgr==2 ? modech.xgr=0 : modech.xgr++); changed=true; break;

      case k7 : (modech.rot==2 ? modech.rot=0 : modech.rot++); changed=true; break;
      case k8 : (modech.wot==2 ? modech.wot=0 : modech.wot++); changed=true; break;
      case k9 : (modech.xot==2 ? modech.xot=0 : modech.xot++); changed=true; break;

      case kA :
      case kU : (modech.suid==2 ? modech.suid=0 : modech.suid++); changed=true; break;

      case kB :
      case kG : (modech.sgid==2 ? modech.sgid=0 : modech.sgid++); changed=true; break;

      case kC :
      case kS : (modech.sticky==2 ? modech.sticky=0 : modech.sticky++); changed=true; break;

      case kD : {
        modech.xot=modech.xgr=modech.xow=_set;
        changed=true;
        break;
      }

      case kE : {
        modech.wot=modech.wgr=modech.wow=_set;
        changed=true;
        break;
      }

      case kF : {
        modech.wot=modech.wgr=modech.wow=_clear;
        changed=true;
        break;
      }
      
      case kF10 :
      case kESC : return false;

      case kEnter : return changed;
      
      default : ;
    }
  } while (!KeyIn(k,kF10,kESC,kEnter));
  return true;
}
  

#ifdef _TARGETFLAVOR_DOS_
//Dialog to allow users to change attributes
bool cActionProject::SetAttrChanges() {
  attrch.archive=attrch.hidden=attrch.system=attrch.readonly=_nochange;
  int i=1;
  do {
    cMenu m("Attribute changes");
    m+=("~Hidden  : "+as2str(attrch.hidden));
    m+=("~ReadOnly: "+as2str(attrch.readonly));
    m+=("~System  : "+as2str(attrch.system));
    m+=("~Archive : "+as2str(attrch.archive));
    m+="~Done";
    m+="a~Bort";
    i=m.Run(i);
    switch (i) {
      case 1 : attrch.hidden++;   if (attrch.hidden>2) attrch.hidden=0; break;
      case 2 : attrch.readonly++; if (attrch.readonly>2) attrch.readonly=0; break;
      case 3 : attrch.system++;   if (attrch.system>2) attrch.system=0; break;
      case 4 : attrch.archive++;  if (attrch.archive>2) attrch.archive=0; break;
      case 5 : return true;
      default : return false;
    }
  } while (i!=0 && i!=5);
  return true;
}
#endif

//Configures the action, like determining destination directory, setup
//attribute/mode changes etc.
bool cActionProject::SetupAction() {
  switch (atype) {
    case acAttr : {
      #ifdef _TARGETFLAVOR_UNIX_
        //Everything is using mode when running Unix
        return SetModeChanges();
      #else
        //Handle that ftpsites use mode eventhough OS is DOS-flavor
        if (FSysUsesMode(fromdir)) {
          SetModeChanges();
        } else {
          SetAttrChanges();
        }
      #endif
    }
    case acCopy : destdir=FmTree->GetPath("Copy to:",fromdir,lastdestination,true); break;
    case acMove : destdir=FmTree->GetPath("Move to:",fromdir,lastdestination,true); break;
    case acUnpack : destdir=FmTree->GetPath("Unpack to:",fromdir,lastdestination,true); break;
    default : destdir="";
  }
  if (destdir==_PRINTERTOKEN) {
    if (atype!=acCopy) {
      ErrorBox("Sorry, that does not make any sense...");
      return false;
    } else return true;
    
  }
  if (destdir=="") {
    //erasing does not need destination dir
    if (atype==acErase || atype==acAttr) return true; else return false;
  } else {
    lastdestination=destdir;
    destdir=AddSlash(destdir);
    if (atype==acMove || atype==acCopy) {
      if (destdir==fromdir) {
        ErrorBox("Cannot copy/move onto itself!");
        return false;
      }
    }
    return true;
  }
}


void cActionProject::UpdateList(int current) {
  //Write files
  static int centerline=((w->GetHeight()-2)/2);
  int curline;
  int curfile;
  if (current<centerline) {
    curline=centerline-current+1; curfile=0;
  } else {
    curline=2; curfile=current-centerline+1;
  }

  while (curline<(w->GetHeight()-2) && curfile<actions.Size()) {
    //Shorten path to look nice
    string s=FixStr(actions[curfile]->GetShowName(),w->GetWidth()-10);
    if (actions[curfile]->GetResult()=="") {
      s+=GetSpaces(w->GetWidth()-s.length()-2);
    } else {
      s+=GetSpaces(w->GetWidth()-s.length()-2-actions[curfile]->GetResult().length());
      s+=actions[curfile]->GetResult();
    }
    if (curfile==current) {
      w->FastWrite(1,curline,s,chWHITE,cRED);
    } else {
      w->FastWrite(1,curline,s);
    }
    curfile++; curline++;
  }

  //Fill empty space on bottom
  if (curline<w->GetHeight()-2) {
    for (int i=curline; i<w->GetHeight()-2; i++) {
      w->FastWrite(1,i,GetSpaces(w->GetWidth()-2));
    }
  }
}

bool cActionProject::Doit() {
  switch (atype) {
    #ifdef _TARGETFLAVOR_UNIX_
      case acAttr   : w->ChTitle("Changing mode on files..."); break;
    #else
      case acAttr   : w->ChTitle("Changing attributes on files..."); break;
    #endif
    case acErase  : w->ChTitle("Erasing files..."); break;
    case acMove   : w->ChTitle("Moving files..."); break;
    case acUnpack : w->ChTitle("Unpacking files..."); break;
    default       : w->ChTitle("Copying files...");
  }
  if (atype==acMove || atype==acCopy) {
    if (!AskSpaceAction() && destdir!=_PRINTERTOKEN) return false;
  }
  totalfilesize=totalmeasure; partfilesize=0;
  totalfilecount=actions.Size();
  w->RedrawBox(); timer.Reset();
  for(int i=0; i<actions.Size(); i++) {
    string s=(MkStr(i+1)+"/"+MkStr(actions.Size())+" files");
    w->ChUpright(s); w->DrawHeader();
    UpdateList(i);
    if (!ShouldI()) {
      actions[i]->SetIgnore();
    } else {
      if (!actions[i]->Action()) {
        actions[i]->SetFailure();
        failures++;
        if (howto_failure==oaUndetermined) {
          AskFailureAction();
        }
        switch (howto_failure) {
          //never ask again
          case oaIgnoreAll : break;

          //ask again next time
          case oaIgnore : howto_failure=oaUndetermined; break;

          //Ask again next time try same file again
          case oaRetry  : {
            howto_failure=oaUndetermined;
            i--;
            //filemeasure-=actions[i].GetSize(); //make "room" for retry
            actions[i]->ClearResult(); failures--; //clear all results
            break;
          }

          //cancel pressed
          default : breaked=true; return true;
        }
      } else actions[i]->SetSuccess();
    };
    if (allpressed) {
      //Any key aborts operation when in ALL-mode
      breaked=IsKeyPressed();
    }
    if (breaked) {
      failures=0; //do not show number of fails if aborted
      return true;
    }
    if (_progress_by_count) {
      partfilesize+=actions[i]->GetSize();
      s=(MkStr(partfilesize,true)+" bytes of "+MkStr(totalfilesize,true));
      w->FastWrite(1,1,s,chYELLOW);
      s=GetProgressBar(i, actions.Size(), w->GetWidth()-2);
      w->FastWrite(1,w->GetHeight()-1,s,chYELLOW);
    }    
  }
  return true;
}

bool cActionProject::ShouldI() {
  if (allpressed) return true;
  switch (GetThatKey(kEnter,kF10,kQ,kESC,kA,kN,kY)) {
    case kN : return false;
    case kA : allpressed=true; return true;
    case kQ :
    case kF10 :
    case kESC : breaked=true; return false;
    default : return true;
  }
}


bool cActionProject::ScanDir(const string& dir) {
  tDirContainer fl;
  if (ScanForFiles(fl,dir,false)) {
    RegisterFilelist(fl,false);
  }
}

void cActionProject::RegisterFilelist(const tDirContainer& fl, bool onlytagged=true) {
  for (int i=0; i<fl.Size(); i++) {
    if (error || breaked) return;
    if (IsKeyPressed()) {
      if (GetSomeKey()==kESC) {error=true; return;};
    }
    cDirEntry de=fl.Read(i);
    if (de.isspecial || de.name=="..") continue;
    if (onlytagged && !de.tagged) continue;

    #ifdef _SYMLINKS_EXIST_
    if (de.islink) {
      if (howto_symlinks==oaUndetermined) AskSymlinkAction();
      if (howto_symlinks==oaIgnore) continue;
      if (howto_symlinks==oaProcess) {
        de.islink=false;
        RegisterFile(de,atype);
        continue;
      }
      //else howto_symlinks==oaResolve -> recurse dirs etc.
    }
    #endif
    if (de.isdir) {
      if (howto_subdirs==oaUndetermined) AskDirAction();
      if (howto_subdirs==oaAlways) { //Let's recurse like hell ;-)
        switch (atype) {
          case acErase : break;
          case acMove : RegisterFile(de,acCopy); break;
          default : RegisterFile(de);
        }
        if (fromdir=="") {
          fromdir=(GetPartDir(de.fullname)+"/");
          ScanDir(de.fullname);
          fromdir="";
        } else {
          ScanDir(de.fullname);
        }
        switch (atype) {
          case acErase :
          case acMove : RegisterFile(de,acErase); break;
          default : {;};
        }
      }
      //hint: howto_subdirs==oaNever -> don't do nothing!
    } else {
      RegisterFile(de); //cAction can now move
    }
  }
}

void cActionProject::RegisterFile(const cDirEntry& de,tActionType type=acDefault) {
  static int line=1;
  string s;
  if (fromdir!="") {
    //only show from current dir and subdirs, not from root (full path)
    if (!IsDirDir(fromdir)) {
      s=de.fullname.substr(fromdir.length()-1,999); //Fix for missing first letter in names
    } else {
      s=de.fullname.substr(fromdir.length(),999);
    }
  } else {
    s=GetPartName(de.fullname);
  }
  if (destdir!="") {
    //check for overwrite!!!!
    if (howto_overwrite==oaUndetermined && DoesExist(destdir+s)) {
      AskOverwriteAction();
    }
    //Cannot reach here with undertermined overwriteaction
    if (howto_overwrite==oaCancel && DoesExist(destdir+s)) {
      error=true; breaked=true;
      return; //Never overwrite
    }

    //Cannot reach here with undertermined overwriteaction
    if (howto_overwrite==oaNever && DoesExist(destdir+s)) {
      return; //Never overwrite
    }
  }
  if (type==acDefault) type=atype;
  switch (type) {
    case acCopy : actions+=new cCopyAction(de,s,(destdir+s),howto_overwrite,&ShowProgress); break;
    case acMove : actions+=new cMoveAction(de,s,(destdir+s),howto_overwrite,&ShowProgress); break;
    case acErase: actions+=new cEraseAction(de,s,&ShowProgress); break;
    case acUnpack: actions+=new cUnpackAction(de,s,destdir,&ShowProgress); break;
    case acAttr : {
      if (de.useattr) {
        actions+=new cChmodeAction(de,s,attrch,&ShowProgress); break;
      } else {
        actions+=new cChmodeAction(de,s,modech,&ShowProgress); break;
      }
    }
    default : {};
  }
  w->FastWrite(1,line,RCopy(s,w->GetWidth()-2,' '));
  filecount++; totalmeasure+=de.size; line++;
  if (line>=w->GetHeight()) line=1;
}

void cActionProject::AskDirAction() {
  if (atype==acUnpack) {
    howto_subdirs=oaNever;
    return;
  }
  //If set in configfile assume always recures unless you are doing modechanges
  if (Configuration.alwaysrecursesubdirs && atype!=acAttr) {
    howto_subdirs=oaAlways;
    return;
  }
  switch (PromptBox("Subdirs encountered!","How should I handle these? ~Recurse/Ignore",kESC,kR,kI,kC,kEnter)) {
    case kEnter :
    case kR : howto_subdirs=oaAlways; break;
    case kI : howto_subdirs=oaNever; break;
    case kC :
    default : howto_subdirs=oaCancel; break;
  }
}

bool cActionProject::AskSpaceAction() {
  if (!IsDirDir(destdir)) return true; //You don't want to check for space in archives/on ftpsites

  string s=destdir; s+='.';
  struct statfs fs;
  if (!statfs(s.c_str(), &fs)) {
    unsigned long int freebytes=(fs.f_bavail*fs.f_bsize);
    if (freebytes>=totalmeasure) return true; //Theres no problem
  } else return false; //Something wrong with the destination

  switch (PromptBox("Not enough diskspace!","What should I do? ~Abort/Continue",kESC,kA,kC,kEnter)) {
    case kC : return true;  //Try anyway
    default : return false; //Give it up
  }
}


void cActionProject::AskOverwriteAction() {
  if (atype==acErase || atype==acAttr) {
    howto_overwrite=oaNever;
    return;
  }
  switch (PromptBox("Existing files encountered!","When should I overwrite these? Always/Never/ne~Wer",kESC,kA,kN,kW,kEnter)) {
    case kA : howto_overwrite=oaAlways; break;
    case kN : howto_overwrite=oaNever; break;
    case kESC : howto_overwrite=oaCancel; break;
    case kEnter :
    default : howto_overwrite=oaNewer;
  }
}

void cActionProject::AskFailureAction() {
  switch (PromptBox("Action failed!","What should I do? ~Ignore/ignoreAll/Retry/Cancel",kESC,kI,kA,kR,kC,kEnter)) {
    case kEnter :
    case kI : howto_failure=oaIgnore; break;
    case kR : howto_failure=oaRetry; break;
    case kA : howto_failure=oaIgnoreAll; break;
    case kESC :
    default : howto_failure=oaCancel;
  }
}

#ifdef _SYMLINKS_EXIST_
void cActionProject::AskSymlinkAction() {
  if (atype==acAttr) {
    howto_overwrite=oaNever;
    return;
  }
  tKey k;
  do {
    k=PromptBox("Symlinks encountered!","What should I do? ~Process/Resolve/Ignore/cAncel/Help",kESC,kH,kF1,kP,kR,kI,kA,kEnter);
    switch (k) {
      case kH :
      case kF1 : ShowSymlinksHelp(); break;
      case kEnter :
      case kP : howto_symlinks=oaProcess; break;
      case kR : howto_symlinks=oaResolve; break;
      case kI : howto_symlinks=oaIgnore; break;
      default : howto_symlinks=oaCancel;
    }
  } while (KeyIn(k,kH,kF1));
}
#endif

/************************************************************************/

void ShowProgress(long int thispart, long int thistotal, int justdone) {
  string s;
  //Write info in top of window
  if (!_progress_by_count) {
    partfilesize+=justdone;
    s=(MkStr(partfilesize,true)+" bytes of "+MkStr(totalfilesize,true));
    win->Write(1,1,s,chYELLOW);
    s=GetProgressBar(partfilesize, totalfilesize, win->GetWidth()-2);
    win->Write(1,win->GetHeight()-1,s,chYELLOW);
  }

  static int lastsec=99;
  if ((lastsec!=timer.Elapsed()) && (partfilesize!=0)) {
    lastsec=timer.Elapsed();
    double bps=((partfilesize*1.0)/timer.Elapsed());
    string speed,etc;
    int secetc=(int)((totalfilesize-partfilesize) / bps);
    if (bps>(1024*1024)) {
      bps/=(1024*1024);
      speed=(MkStr(bps)+" mb/sec");
    } else {
      if (bps>(1024)) {
        bps/=1024;
        speed=(MkStr(bps)+" kb/sec");
      } else {
        int i=(int)bps;
        speed=(MkStr(i)+" bts/sec");
      }
    }
    win->ChDnleft(speed);
    etc="ETC: ";
    if (secetc>60) {
      etc=(MkStr(secetc/60)+" min. ");
      secetc=secetc % 60;
    }
    etc+=(MkStr(secetc)+" sec.");
    win->ChDnright(etc);
    win->DrawHeader();
  }

  //Write progress bar in bottom
  s=GetProgressBar(thispart, thistotal, win->GetWidth()-2);
  win->Write(1,win->GetHeight()-2,s,chYELLOW);
}
