#include "target.h"
#ifdef _TARGET_LINUX_
//  #include <sys/io.h>
#else
  #include <io.h>
#endif
#include <stdlib.h>
#include "sc_file.h"
#include "fm_vfs.h"
#include "fm_window.h"
#include "sc_archive.h"
#include "sc_ftp.h"
#ifdef _BZ2_SUPPORT_
  #include <bzlib.h>
#endif

//Taken from gzio.h from zlib-package, for typecasting the gzFile (void*)
//to this struct. Hereby I can access the actual FILE*, and determine
//pos and size for progress calculation. Is there a nicer way?
//gzFile has a gztell-routine, but no gzsize-routine (which would of course
//be darn slow, if it was implemented)
typedef struct gz_stream {
    z_stream stream;
    int      z_err;   /* error code for last stream operation */
    int      z_eof;   /* set if end of input file */
    FILE     *file;   /* .gz file */
    Byte     *inbuf;  /* input buffer */
    Byte     *outbuf; /* output buffer */
    uLong    crc;     /* crc32 of uncompressed data */
    char     *msg;    /* error message */
    char     *path;   /* path name for debugging only */
    int      transparent; /* 1 if input file is not a .gz file */
    char     mode;    /* 'w' or 'r' */
    long     startpos; /* start of compressed data in file (header skipped) */
} gz_stream;



class cDiskFile:public cFile {
  public:
    //******************
    cDiskFile(const string& name) : cFile(name) {};

    //******************
    ~cDiskFile() {
      if (isopen) Close();
    };

    //********************
    bool Open(tOpenMode om) {
      mode=om;
      switch (mode) {
        //Open as gzfile, if it is not data are simply passed through
        #ifdef _BZ2_SUPPORT_
          case omGZRead  : if (GetPartExt(filename)==".TAR.BZ2") {
            bzf=BZ2_bzopen(filename.c_str(), "rb"); isopen=(bzf!=NULL);
            bznow=true; break;
          } else {
            gzf=(gzFile*)gzopen(filename.c_str(), "rb"); isopen=(gzf!=NULL);
            bznow=false; break;
          }
        #else
          case omGZRead  : gzf=(gzFile*)gzopen(filename.c_str(), "rb"); isopen=(gzf!=NULL); break;
        #endif

        //Open as normal file, do not uncompress GZ-data
        case omRead : f=fopen(filename.c_str(), "rb"); isopen=(f!=NULL); break;

        //Open as noncompressed file for writing
        case omWrite : f=fopen(filename.c_str(), "wb"); isopen=(f!=NULL); break;

        //Open as compressed file for writing
        case omGZWrite : gzf=(gzFile*)gzopen(filename.c_str(), "wb"); isopen=(gzf!=NULL); break;
      }
      return (isopen);
    }

    //********************
    int Read(void* buf, int size) {
      if (!isopen) return -1;
      switch (mode) {
        #ifdef _BZ2_SUPPORT_
          case omGZRead : if (bznow) {
            return BZ2_bzread(bzf,buf,size);
          } else {
            return gzread(gzf,buf,size); //noncompresed filedata passed through
          }
        #else
          case omGZRead :return gzread(gzf,buf,size); //noncompresed filedata passed through
        #endif
        case omRead : return fread(buf,1,size,f);
        default : return -1; //Did you try to write a read-file?
      }
    }

    //*********************
    int Write(void* buf, int size) {
      if (!isopen) return -1;
      switch (mode) {
        case omWrite   : return fwrite(buf,1,size,f);
        case omGZWrite : return gzwrite(gzf,buf,size);
        default : return -1; //Did you try to read a write-file?
      }
    }

    //*********************
    void Close() {
      if (!isopen) return;
      switch (mode) {
        case omRead :
        case omWrite : fclose(f); break;
        #ifdef _BZ2_SUPPORT_
          default : if (bznow) BZ2_bzclose(bzf); else gzclose(gzf);
        #else
          default: gzclose(gzf);
        #endif
      }


      isopen=false;
    }

    //*********************
    bool Delete() {
      if (isopen) {
        return false;
      } else {
        return DeleteFile(filename);
      }
    }

    //**********************
    int Size() {
      if (isopen) {
        //#ifdef _TARGET_LINUX_
          return GetFileSize(filename);
        /*#else
        switch (mode) {
          case omRead :
          case omWrite : return filelength(fileno(f)); break;

          //This is why I think c/c++ code can be _really_ ugly!
          default : return filelength(fileno( ((gz_stream*)gzf)->file ));
        }
        #endif*/
      } else {
        return GetFileSize(filename);
      }
    }

    //***********************
    int Pos() {
      if (isopen) {
        switch (mode) {
          case omRead :
          case omWrite : return ftell(f);
      
          //This is why I think c/c++ code can be _really_ ugly!
          #ifdef _BZ2_SUPPORT_
            default : if (bznow) {
              return 0;
            } else {
              return ftell( ((gz_stream*)gzf)->file );
            }
          #else
            default : return ftell( ((gz_stream*)gzf)->file );
          #endif
        }
      } else {
        return -1;
      }
    }

    //***********************
    int GetMode() {
      #ifdef _TARGET_DOS_
        return _chmod(filename.c_str(),0);
      #else
      if (isopen) {
        struct stat s;
        if (stat(filename.c_str(), &s)==0) {
          return s.st_mode;
        } else return -1;
      } else {
        struct stat s;
        if (fstat(fileno(f), &s)==0) {
          return s.st_mode;
        } else return -1;
      }
     #endif
    }

    //***********************
    bool SetMode(int m) {
      #ifdef _TARGET_DJGPP_
        return (_chmod(filename.c_str(),1,m)!=-1);
      #else
        return (chmod(filename.c_str(),m)==0);
      #endif
    }
  private:
    union {
      FILE* f;
      gzFile* gzf;
      #ifdef _BZ2_SUPPORT_
        BZFILE* bzf;
      #endif
    };
    #ifdef _BZ2_SUPPORT_
      bool bznow;
    #endif
/*    bool isopen;
    string filename;*/
};

/*****************************************************************************/

cFile* GetFilePointer(const string& dir) {
  sParsedDir spd=PleaseParseDir(dir);
  if (spd.type==dtArc) {
    return GetArcFilePointer(spd);
  }
  #ifdef _FTP_POSSIBLE_
  if (spd.type==dtFtp) {
    return GetFtpFilePointer(spd);
  }
  #endif
  return new cDiskFile(dir);
}

void DisposeFilePointer(cFile* f1,cFile* f2/*=NULL*/) {
  f1->Close(); delete f1;
  if (f2!=NULL) {
    f2->Close(); delete f2;
  }
}


bool CopyFile(const string& filename1, const string& filename2,bool progress/*=false*/) {
  //cFile f1(filename1);
  cFile* f1=GetFilePointer(filename1);
  if (





  !f1->Open(omRead)) {delete f1; return false;};

  //cFile f2(filename2);
  cFile* f2=GetFilePointer(filename2);
  if (!f2->Open(omWrite)) {delete f1; delete f2; return false;}
  //f1 is closed ind destructor if open

  char buf[65536];
  int i=f1->Read(&buf,sizeof(buf));
  while (i>0) {
    f2->Write(&buf,i);
    i=f1->Read(&buf,sizeof(buf));
    ShowProgressWindow(f1->Pos(),f1->Size(),50,"Duplicating...");
  }
  /*f1->Close(); f2->Close();
  delete f1; delete f2;*/
  DisposeFilePointer(f1,f2);
  return true;
}

int GetChunkOfFile(const string& filename, int size, char buffer[]) {
/*  FILE *f=fopen(filename.c_str(), "rb+");
  if (!f) return false;
  int read=fread(buffer,1,size,f);
  fclose(f);
  return read;*/

//Of some reason using this crashes a subsequent read when it's finished.
//Why?
  cFile* f=GetFilePointer(filename);
  if (!f->Open(omRead)) return -1;
  int i=f->Read(buffer,size);
  f->Close();
  delete f;

  return i; //Automatically close in destructor
}

bool IsFileBinary(const string& filename) {
  char buf[512];
  if(int i=GetChunkOfFile(filename,sizeof(buf),buf)) {
    for(int n=0; n<i-1; n++) {
      if (buf[n]=='\0') {
        return true;
      }
    }
  }
  return false;
}

